;DIY-UHF Digital encoded signal RX
;David Lancaster
;All rights reserved

;v2.2
;	13pf XTAL impedance to match CTX658 Xtal
;	Cleaned up Frame/Block definition
;	Removed unused interrupt code
;	Streamlined parsing input and PFM code
;	Added FHSS code
;	Bug fixes
;	Stopped using nRIQ, testing IA_SDO instead
;	Rebranded as version 2 for simplicity
;v3.1
;	Added dwell counter to reduce number of actual channel changes	
;	Switched to 8-bit channel values

;CONFIGURABLE OPTIONS

#define cERROR_TIMEOUT			d'600000'	;approximate number of uS we'll fly on a dropped signal before shutting down												


;END CONFIGURABLE OPTIONS


	MESSG "Assembling for 16F630"
	LIST	P=PIC16F630
	INCLUDE "P16F630.INC"
	__CONFIG _MCLRE_OFF & _CP_OFF & _WDT_OFF & _INTRC_OSC_NOCLKOUT & _CPD_OFF & _PWRTE_ON & _BODEN
	;ERRORLEVEL -302					;turn off register not in bank0 warnings


#INCLUDE ../data_format.inc


#define	cSIGNAL_USE_THRESHOLD	cERROR_TIMEOUT/((cPFM_FREQ)*d'255')
#define	cPFM_FREQ				d'256'-d'50'		;min limit on PFM rate
#define cPFM_ON_TIME			d'24'
#define	cHOPTIMEOUT				d'100000'			;number of uS to wait on a channel for a binding TX

#define NEW_INPUTS		MISC,0
#define LOST_SIGNAL		MISC,1
#define	TEMPA			MISC,2
#define	TEMPB			MISC,3
#define	TEMPTOGGLE		MISC,4
#define	LOW_BATT		MISC,5	;flags low battery or not sync'ed (at power-up)
#define	BOUND_TO_TX		MISC,6	;indicates listening for specified TX code


;PFM_FLAGS
#define	AILTOGGLE	PFM_FLAGS,0
#define	ELETOGGLE	PFM_FLAGS,1
#define	ESCTOGGLE	PFM_FLAGS,2
#define	RUDTOGGLE	PFM_FLAGS,3


;PIN LOCATION SPECIFIC settings
;BE EXTREMELY CAREFUL WITH THESE!

;DIY-UHF RX4 16F630 RX pinouts:
;
;		   	---------
;		Vcc	|1    14| Gnd
;nSEL<--RA5	|2    13| RA0--> SDI
;nIRQ-->RA4 |3    12| RA1--> SCK
;SDO -->RA3	|4    11| RA2--> ESC
;ELEB<--RC5 |5    10| RC0--> RUDA
;ELEA<--RC4 |6     9| RC1--> RUDB
;AILB<--RC3 |7     8| RC2--> AILA
;		   	---------

#define 	TRIS_PORTA	b'00011000'			;PORTA all output, except RA3,RA4
#define		TRIS_PORTC	b'00000000'			;PORTC all outputs

;pin connections to IA 4320 RX chip
#define		IA_SDI		PORTA,0
#define		IA_SCK		PORTA,1

#define		IA_SDO		PORTA,3
#define		IA_nIRQ		PORTA,4
#define		IA_nSEL		PORTA,5



;position of channel pins
#define	RUDPINAPOS		0
#define	RUDPINBPOS		1
#define	AILPINAPOS		2
#define	AILPINBPOS		3
#define	ELEPINAPOS		4
#define	ELEPINBPOS		5

#define	ESCPINPOS		2	;<-Note, PORTA, not PORTC




;IA4320 settings
;these require checking the datasheet, have fun
#define		IA_CONFIG		b'1001110010011001'			;915 Mhz band, low battery detect on, no wake-up timer or osc in sleep
														;13 pf XTL capacitance, 200khz baseband bw, disable clockout
#define		IA_FREQ			b'1010011111010000'			;915.00 Mhz
#define		IA_RX			b'1100000001000000'		 	;VDI=DQD, Max LNA gain, RX off
#define		IA_LBD			b'1100001000001000'			;set for 3.2V low battery detect (3.0V + ~0.2V offset on PCB), no clkout
#define		IA_AFC			b'1100011000000111'			;disable AFC, unlimited offset, fine mode
#define		IA_DATAFILT		b'1100010001101100'			;fast clock lock, Digital filtering, DQD=4
#define		IA_DATARATE		b'1100100000100011'			;set expected bit rate as ~9600 (9579bps)
#define		IA_FIFO			b'1100111010000101'			;FFIT @ 8bits, start fill on sync word, disable fill, enable fifo

#define		IA_FIFO_ENABLE	b'00000010'					;used to enable FIFO when or'ed with IA_FIFO
#define		IA_RX_ENABLE	b'00000001'					;used to enable RX when or'ed with IA_RX


;Define memory register aliases
 CBLOCK 0x020

	;Values received from TX, these *must* be adjacent in ascending order (addressed by FSR/INDF later)
	FRAME0
	FRAME1
	FRAME2
	FRAME3
	FRAME4
	FRAME5
	FRAME6
	INPUT_WORD_COUNT

	ID_CODE
	CURRENTCHAN

	;buffers for data once we've had a successful read, to minimize the chance of "clobbering"
	AIL_INPUT
	ELE_INPUT
	ESC_INPUT
	RUD_INPUT

	;values for PFM
	AIL_PFM_VAL								;PFM set values
	ELE_PFM_VAL
	ESC_PFM_VAL
	RUD_PFM_VAL
	AIL_PIN_VAL								;settings for output pins
	ELE_PIN_VAL
    ESC_PIN_VAL
	RUD_PIN_VAL
	AIL_PFM_CNTR							;PFM counters
	ELE_PFM_CNTR
	ESC_PFM_CNTR
	RUD_PFM_CNTR
	PFM_FLAGS

	SIGNAL_USE_COUNTER						;counters to track how often we re-use a signal
	SIGNAL_USE_COUNTER2
	TEMP

	;SPI regs
	IA_DATAH
	IA_DATAL
	SPI_BIT_COUNTER
	SPI_TEMP

	MISC
 ENDC


;define Macros
;these are short enough to be more efficent as macros than subroutines
goBank1 MACRO
  bsf STATUS,RP0
  ERRORLEVEL -302							;acknowledge that we're in bank1, stop bank error warnings
 ENDM

goBank0 MACRO
  bcf STATUS,RP0
  ERRORLEVEL +302							;clear bank1 ack so we see banking error warnings again
 ENDM



;PROGRAM CODE STARTS HERE
 ORG 0  
  goto SETUP

;-------------------------------------
HANDLEINTERRUPT
	;found an interrupt, read out the status bits and figure out what the RX chip is notifying us about
  bcf IA_SCK
  bcf IA_SDI								;read command is to send a CLK with SDI low (all other commands start high)
  bcf IA_nSEL								;nSEL set low to open SPI conversation
  nop										;min SPI wait is 10 ns, 1 uS should do dandy ;)
	;Get the first 8 bits, which include the four interrupt bits (7-FFIT, 6-FFOV, 5-WKUP, 4-LBD)
  call get_8SPI_BITs
  movwf IA_DATAH
  	;test LBD
  bcf LOW_BATT
  btfsc IA_DATAH,4
  bsf LOW_BATT								;flag that the battery is low
  btfss IA_DATAH,6							;check FFOV
  goto NOFFOV
  call FINISH_SPI							;FIFO overflow, Houston we have a problem...reset and leave
  call RESETDATASTREAM						
  return
NOFFOV
  btfss IA_DATAH,7							;check if FFIT is set
  goto FINISH_SPI							;FFIT not set, no data, so stop reading, but no reset
  call get_8SPI_BITs						;FFIT set, we should have 8 bits of data (after reading 8 more status bits)
  movwf IA_DATAL							;8 more status bits
	;Now the juicy bit, real data from the TX
  call get_8SPI_BITs
  movwf SPI_TEMP							;8 bits of data, w00t

	;save data to correct register
  movlw FRAME0								;load first address into indirect address register
  movwf FSR									
  movfw INPUT_WORD_COUNT					;add offset
  addwf FSR,f								
  movfw SPI_TEMP							;save using indirect addressing
  movwf INDF								
  incf INPUT_WORD_COUNT,f					;increment to point to next data word storage location
  ;fall into FINISH_SPI

FINISH_SPI
	;done reading this byte, close off SPI conv
  nop
  bsf IA_nSEL								;finish SPI conversation
  bsf IA_SDI			;LED off by default  
  btfsc BOUND_TO_TX		;set LED if we're not bound
  btfsc LOW_BATT		;set LED if we're low on battery
  bcf IA_SDI			;turn LED on
  movfw INPUT_WORD_COUNT
  sublw FRAME_FIELDS						;test if we've seen the entire frame
  btfsc STATUS,Z							;this means extra data is ignored, could be used for more channels or optional
  goto FINISHED_FRAME						;settings/configuration data
  return


FINISHED_FRAME
	;we received a full frame of data, save it out to more permanent registers
	;test if the frame is good
  andlw d'0'				;clear W
  addwf FRAME0,w			;add up all of the bytes of the FRAME0
  addwf FRAME1,w			; and check against the CHECKSUM
  addwf FRAME2,w
  addwf FRAME3,w
  addwf FRAME4,w
  addwf FRAME5,w
  ;addwf FRAME6,w			;don't include the transmitted checksum in the addition...
  subwf CHECKSUM_VAL,w		;received from TX
  btfsc STATUS,Z
  goto VALID_FRAME
  call RESETDATASTREAM		;reset the FIFO, and counters so we can pickup the next block
  return					;frame corrupted, do nothing with it and stay on this freq till timeout since we can't verify it was from our TX

VALID_FRAME
  call RESETDATASTREAM		;reset the FIFO, and counters so we can pickup the next block
  movfw ID_CODE_VAL			;check if it matches our ID
  subwf ID_CODE,w
  btfsc STATUS,Z
  goto VALID_CODE			;matches, use it
	;else, it's not addressed to us
  btfsc BOUND_TO_TX			;see if we are already bound to a TX
  return					;already bound, ignore the transmission and keep listening on this freq till timeout
  movfw SIGOPTS_VAL		;else we're not bound, see if it's a binding advertisement
  andlw SIGOPTS_TYPEMASK
  sublw ADV_BINDING
  btfss STATUS,Z
  return					;not a binding advertisement, ignore it
  bcf T1CON,TMR1ON			;turn off TMR1 since we don't need to jump channels anymore
  movfw ID_CODE_VAL			;else, accept it as our controlling TX and synchronize our hopping with it
  movwf ID_CODE
  bsf BOUND_TO_TX
	;Set AFC command
  movlw HIGH(IA_AFC)
  movwf IA_DATAH
  movlw LOW(IA_AFC) | b'00001000'	;strobe the AFC *now*
  movwf IA_DATAL
  movlw d'16'
  call send_SPI   
  ;goto VALID_CODE	;fall through to VALID_CODE


VALID_CODE
  movfw SIGOPTS_VAL			;see if it's a regular data transmission
  andlw SIGOPTS_TYPEMASK
  sublw REG_TX
  btfsc STATUS,Z
  goto REG_TRANSMISSION
	;else, we don't want/know what to do with it, so ignore it, but keep hopping with the TX
  return


REG_TRANSMISSION
  movfw AIL_VAL				;else grab the data and use it
  movwf AIL_INPUT
  movfw ELE_VAL
  movwf ELE_INPUT
  movfw ESC_VAL
  movwf ESC_INPUT
  movfw RUD_VAL
  movwf RUD_INPUT
  bsf NEW_INPUTS
  return


RESETDATASTREAM
	;reset the datastream status, clear the FIFO, and it will look for the synchron pattern again
  clrf INPUT_WORD_COUNT
  bcf NEW_INPUTS
	;make sure any current SPI conversation is over
  bsf IA_nSEL
	;Disable/Clear FIFO
  movlw HIGH(IA_FIFO)
  movwf IA_DATAH
  movlw LOW(IA_FIFO)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
	;Re-enable FIFO
  movlw HIGH(IA_FIFO)
  movwf IA_DATAH
  movlw LOW(IA_FIFO) | IA_FIFO_ENABLE
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  return

;END HANDLEINTERRUPT
;-------------------------------------



;-------------------------------------
;Sub SHIFTCHAN
SHIFTCHAN
	;cycle through the channels

	;Change Frequency Setting based on CURRENTCHAN
  pageselw GETCHANHIGH
  movfw CURRENTCHAN
  call GETCHANHIGH							;lookup high byte of channel
  movwf IA_DATAH	
  pageselw GETCHANLOW
  movfw CURRENTCHAN
  call GETCHANLOW							;lookup low byte of channel
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  
  incf CURRENTCHAN,f
  movfw CURRENTCHAN
  sublw NUM_CHANNELS						;test if CURRENTCHAN > NUM_CHANNELS (CURRENTCHAN is zero indexed)
  btfsc STATUS,Z
  clrf CURRENTCHAN							;reset to the first channel		

	;set TMR1 overflow duration
  movlw d'256'-( (cHOPTIMEOUT/2) % d'256')			
  movwf TMR1L														
  movlw d'256'-( (cHOPTIMEOUT/2) / d'256')
  movwf TMR1H
  bcf PIR1,TMR1IF							;clear TMR1 interrupt flag just in case
  movlw b'00010001'							;config for 1:2 prescale, TMR1 on
  movwf T1CON
  bsf T1CON,TMR1ON
  
  call RESETDATASTREAM						;reset the FIFO, and counters so we can pickup the next frame
  return

;END SHIFTCHAN
;-------------------------------------



;-------------------------------------
get_8SPI_BITs
	;fetch 8 bits
	;Note that IS4230 will assert an extra bit from the FIFO on the last falling clock edge, but it won't purge it
	;This confused me the heck out of me, as it shows up on the oscilloscope...
  movlw d'8'								;8 bits
  movwf SPI_BIT_COUNTER
  clrf SPI_TEMP
  
get_SPI_BIT
  rlf SPI_TEMP,f
  bsf IA_SCK	;SCK rising edge
  bcf SPI_TEMP,0							;pull in single bit off SDO of IA4320
  btfsc IA_SDO		
  bsf SPI_TEMP,0
  bcf IA_SCK	;SCK falling edge
  
  decfsz SPI_BIT_COUNTER,f
  goto get_SPI_BIT
  
  movfw SPI_TEMP
  
  return
;-------------------------------------



;-------------------------------------
;Sub send_SPI
send_SPI
	;Sub to send specified # (in W) of bits via SPI (up to 16) from registers IA_DATAL & IA_DATAH
	;Starts with MSB of IA_DATAH, so bits should be left justified to IA_DATAH
  movwf SPI_BIT_COUNTER
  bcf IA_SCK
  bcf IA_SDI
  bcf IA_nSEL	;nSEL set low to open SPI conversation
  nop
send_SPI_BIT
  bcf IA_SDI	;set value on SD output (SDI on IA chip)
  btfsc IA_DATAH,7
  bsf IA_SDI
  nop
  rlf IA_DATAL,f
  bsf IA_SCK	;SCK rising edge (note, does not affect bit in STATUS,C from rlf)
  rlf IA_DATAH,f
  bcf IA_SCK	;SCK falling edge
  decfsz SPI_BIT_COUNTER,f
  goto send_SPI_BIT
;  goto $+1
  bsf IA_nSEL	;finish SPI conversation

	;the LED is lit by setting the SDI output low when nSEL is high (no SPI conversation)
	; nSEL --|>|--\/\/\--- SDI
	;         \\    
  bsf IA_SDI			;LED off by default
  btfsc BOUND_TO_TX		;set LED if we're not bound
  btfsc LOW_BATT		;set LED if we're low on battery
  bcf IA_SDI			;turn LED on

  return
;END send_SPI
;-------------------------------------


;-------------------------------------
WAIT256_W
	;this will wait 256 uS x W
  bcf INTCON,T0IF
  clrf TMR0									;TMR0 will overflow after W increments
  subwf TMR0,f
  goBank1
  movlw b'00000111'							;TMR0 config:prescaled 1:256
  movwf OPTION_REG
  goBank0
  btfss INTCON,T0IF	
  goto $-1
  bcf INTCON,T0IF
  return
;-------------------------------------


;-------------------------------------
SETUP
  clrf PORTA 								;clear outputs
  clrf PORTC								;clear outputs
  goBank1
  call 3FFh									
  movwf OSCCAL 								;config OSCCAL
  movlw b'00001000'							;TMR0 config: no prescaler, internal clocking (Fosc/4 = 1 cycle)
  movwf OPTION_REG
  clrf PIE1
  movlw TRIS_PORTA
  movwf TRISA							 
  movlw TRIS_PORTC							
  movwf TRISC
  goBank0
  movlw b'00000001'							;TMR1 config: no prescaler, internal clocking (Fosc/4 = 1 cycle)
  movwf T1CON
  bcf T1CON,TMR1ON							;disable TMR1
  movlw 07h									;Turn off Comparator unit (111)
  movwf CMCON


;done PIC setup


;init protocol
  clrf INTCON								;clear interrupt settings
  bcf NEW_INPUTS							;clear flags
  clrf MISC
  bsf LOST_SIGNAL							;start in a "lost signal" fashion
  clrf ID_CODE
  clrf CURRENTCHAN							;start on the first channel
;done protocol setup


;do IA4320 setup
  bsf IA_nSEL		;it's logic low, so set it high

;sleep for a bit so the RX chip can stabilize after it's POR
  movlw d'1000000'/d'65536'					;one second, rounded down
  movwf TEMP
OUTER_LOOP
  movlw d'255'
  call WAIT256_W
  decfsz TEMP,f
  goto OUTER_LOOP

	;Receiver Setting *shut off RX*
  movlw HIGH(IA_RX)
  movwf IA_DATAH
  movlw LOW(IA_RX)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W
	;Configuration Setting
  movlw HIGH(IA_CONFIG)
  movwf IA_DATAH
  movlw LOW(IA_CONFIG)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI 
  movlw d'2'
  call WAIT256_W
	;Frequency Setting
  movlw HIGH(CHANNEL1)
  movwf IA_DATAH
  movlw LOW(CHANNEL1)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W
	;Wake up timer left at default
	;Low duty cycle left at default
	;Low battery threshold & clockout
  movlw HIGH(IA_LBD)
  movwf IA_DATAH
  movlw LOW(IA_LBD)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W
	;Data Filter
  movlw HIGH(IA_DATAFILT)
  movwf IA_DATAH
  movlw LOW(IA_DATAFILT)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W
	;Data Rate setting
  movlw HIGH(IA_DATARATE)
  movwf IA_DATAH
  movlw LOW(IA_DATARATE)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W
	;AFC command
  movlw HIGH(IA_AFC)
  movwf IA_DATAH
  movlw LOW(IA_AFC)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W
	;Receiver Setting & enable RX
  movlw HIGH(IA_RX)
  movwf IA_DATAH
  movlw LOW(IA_RX) | IA_RX_ENABLE
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W
	;Output/FIFO setting
  movlw HIGH(IA_FIFO)
  movwf IA_DATAH
  movlw LOW(IA_FIFO)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  movlw d'2'
  call WAIT256_W

	;Wait for Oscillator & PLL to stabilize
  movlw d'10'
  call WAIT256_W

	;Enable FIFO now
  movlw HIGH(IA_FIFO)
  movwf IA_DATAH
  movlw LOW(IA_FIFO) | IA_FIFO_ENABLE
  movwf IA_DATAL
  movlw d'16'
  call send_SPI


  goBank1
  movlw b'00001110'							;TMR0 config: no prescaler, Prescale Watchdog by 1:64
  movwf OPTION_REG
  goBank0
  movlw cPFM_FREQ
  movwf TMR0								;set TMR0 for handling PFM

  ;bsf T1CON,TMR1ON							;set TMR1 for handling SHIFTCHAN timeouts
  call SHIFTCHAN

;-------------------------------------
MAIN
	;main body of the code
	;we loop in here forever

  ;poll interrupt pin (active low)
  btfss IA_nIRQ
  ;poll SDO pin to see if FFIT occured (active high)
  ;btfsc IA_SDO
  call HANDLEINTERRUPT

	;test if we timed out waiting for a binding TX
  btfsc PIR1,TMR1IF
  call SHIFTCHAN

;testdata for testing PFM
; MESSG "Test data deployed, this code won't actually work!"
;  movlw b'00101000'	
;  movwf AIL_INPUT
;  movlw b'01110000'	
;  movwf ELE_INPUT
;  movlw b'00011100'	
;  movwf ESC_INPUT
;  movlw b'00001000'	
;  movwf RUD_INPUT

;
;  btfsc BOUND_TO_TX
;  bsf PORTC,0
;  btfss NEW_INPUTS
;  goto MAIN
;  movlw d'64'
;  movwf TEMP
;LOOP
;  bcf PORTC,1
;  btfsc FRAME0,7
;  bsf PORTC,1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  goto $+1
;  bcf STATUS,C
;  rlf FRAME6,f
;  rlf FRAME5,f
;  rlf FRAME4,f
;  rlf FRAME3,f
;  rlf FRAME2,f
;  rlf FRAME1,f
;  rlf FRAME0,f
;  decfsz TEMP,f
;  goto LOOP
;  bcf NEW_INPUTS
;  bcf PORTC,1
;
  
  
  
  

  btfsc NEW_INPUTS
  call PARSE_INPUTS

	;we don't use the interrupt on TMR0 overflow, but we can still test on the flag
  btfss INTCON,T0IF
  goto MAIN

	;TMR0 expired, set counter again and do some PFM if we have recent signals
  movlw cPFM_FREQ
  movwf TMR0
  bcf INTCON,T0IF
  btfss LOST_SIGNAL
  call DOPFM								;do PFM on the outputs

  goto MAIN
;END MAIN
;-------------------------------------




;-------------------------------------
PARSE_INPUTS
  bcf NEW_INPUTS

  bcf AILTOGGLE
  clrf AIL_PIN_VAL
  movfw AIL_INPUT
  call PARSEINPUT
  movwf AIL_PFM_VAL
  btfsc TEMPA
  bsf AIL_PIN_VAL, AILPINAPOS
  btfsc TEMPB
  bsf AIL_PIN_VAL, AILPINBPOS
  btfsc TEMPTOGGLE
  bsf AILTOGGLE

PARSE_ELE
  bcf ELETOGGLE
  clrf ELE_PIN_VAL
  movfw ELE_INPUT
  call PARSEINPUT
  movwf ELE_PFM_VAL
  btfsc TEMPA
  bsf ELE_PIN_VAL, ELEPINAPOS
  btfsc TEMPB
  bsf ELE_PIN_VAL, ELEPINBPOS
  btfsc TEMPTOGGLE
  bsf ELETOGGLE

	;esc is a bit different
PARSE_ESC
  bcf ESCTOGGLE
  clrf ESC_PIN_VAL
  bsf ESC_PIN_VAL, ESCPINPOS
  movfw ESC_INPUT
  ;btfsc LOW_BATT				;if battery is low, cut motor
  ;goto ESC_MIN
  sublw d'255'					;pivot value around 127 to get PFM frequency
  movwf ESC_PFM_VAL
  sublw d'5'					;test if it's maxed
  btfsc STATUS,C
  goto PARSE_RUD				;esc maxed, leave it on with toggling disabled
  movfw ESC_PFM_VAL
  sublw d'255'-d'20'			;test if it's low enough to consider off
  btfss STATUS,C
  goto ESC_MIN
  bsf ESCTOGGLE					;esc in normal pfm range, enable pfm toggling
  goto PARSE_RUD
ESC_MIN
  bcf ESC_PIN_VAL, ESCPINPOS	;esc at min, turn off the pin with toggling disabled


PARSE_RUD
  bcf RUDTOGGLE
  clrf RUD_PIN_VAL
  movfw RUD_INPUT
  call PARSEINPUT
  movwf RUD_PFM_VAL
  btfsc TEMPA
  bsf RUD_PIN_VAL, RUDPINAPOS
  btfsc TEMPB
  bsf RUD_PIN_VAL, RUDPINBPOS
  btfsc TEMPTOGGLE
  bsf RUDTOGGLE

	;prime PFM counters
  movlw cPFM_ON_TIME
  movwf AIL_PFM_CNTR
  movwf ELE_PFM_CNTR
  movwf ESC_PFM_CNTR
  movwf RUD_PFM_CNTR

	;now move new pin values to PORTC & ESC_PIN
  movlw 0
  iorwf AIL_PIN_VAL,w
  iorwf RUD_PIN_VAL,w
  iorwf ELE_PIN_VAL,w
  movwf PORTC
	;set the ESC without causing a "hiccup"
  btfsc ESC_PIN_VAL, ESCPINPOS
  goto SETESCHIGH
  bcf PORTA,ESCPINPOS
  goto FLAG_GOOD_INPUT
SETESCHIGH
  bsf PORTA,ESCPINPOS


FLAG_GOOD_INPUT
	;received a set of good inputs, reset the lost-signal counters
  bcf LOST_SIGNAL
  movlw cSIGNAL_USE_THRESHOLD
  movwf SIGNAL_USE_COUNTER
  movlw d'255'
  movwf SIGNAL_USE_COUNTER2

  return
;END PARSE_INPUTS
;-------------------------------------




;-------------------------------------
PARSEINPUT
;convert value from 8-bit [0-255] directional value to [0-127] strength value & direction bit
;including setting pin direction and toggling flags
;expect inputs in W, writes modified value out to W
; and sets/clears flags TEMPPOS, TEMPA, TEMPB
  bcf TEMPA
  bcf TEMPB
  bsf TEMPTOGGLE
  movwf TEMP
  btfss TEMP,7						;test if it's >=128, ie. over half
  goto CH_NEG						;under half, we can use the raw value
	;channel is positive, subtract from 255
  movfw TEMP
  bsf TEMPA
  sublw d'255'
  movwf TEMP
  goto TEST_MARGINS							
CH_NEG 					
  bsf TEMPB
TEST_MARGINS										
  ;now we have a strength only value we can use for PFM
  ;where max is ~0, and min is ~127 (and a flag indicates actual direction)
  ;now test against our max/min bounds (technically, min will be around center)
  sublw d'5'
  btfsc STATUS,C
  bcf TEMPTOGGLE					;max value, turn it on, full time (no toggling)
  movlw d'127'-d'3'
  subwf TEMP,w
  btfss STATUS,C
  goto PARSEEND
  bcf TEMPA							;min value = center-ish, turn it off
  bcf TEMPB
  bcf TEMPTOGGLE
PARSEEND
  movfw TEMP						;move the calculated value back to W
  return
;END PARSEINPUT
;-------------------------------------




;-------------------------------------
DOPFM
  decfsz SIGNAL_USE_COUNTER2,f
  goto NORMALPFM
  decfsz SIGNAL_USE_COUNTER,f
  goto NORMALPFM
	;if we get here, the signal has been over-used without being refreshed
	;go into "lost-signal" mode by clearing the outputs
  bsf LOST_SIGNAL
  clrf PORTC
  bcf PORTA,ESCPINPOS
  return

NORMALPFM
	;strategy with PFM is to test if we need to toggle the channel (min/max stay full off/on)
	;if so, decrement and check the counter, toggling when the counter expires
  btfss AILTOGGLE
  goto TEST_ELE
  decfsz AIL_PFM_CNTR,f
  goto TEST_ELE										;counter still hasn't expired, move on
  movfw AIL_PIN_VAL
  xorwf PORTC,f
  andwf PORTC,w
  movlw cPFM_ON_TIME									;load PFM on time, Note that movlw doesn't affect Z
  btfsc STATUS,Z									;if Z is set, the pins are off, use the off value
  movfw AIL_PFM_VAL									;set cntr for desired time
  movwf AIL_PFM_CNTR

TEST_ELE
  btfss ELETOGGLE
  goto TEST_ESC
  decfsz ELE_PFM_CNTR,f
  goto TEST_ESC										;counter still hasn't expired, move on
  movfw ELE_PIN_VAL
  xorwf PORTC,f
  andwf PORTC,w
  movlw cPFM_ON_TIME									;load PFM on time, Note that movlw doesn't affect Z
  btfsc STATUS,Z									;if Z is set, the pins are off, use the off value
  movfw ELE_PFM_VAL									;set cntr for desired time
  movwf ELE_PFM_CNTR

TEST_ESC
  btfss ESCTOGGLE
  goto TEST_RUD
  decfsz ESC_PFM_CNTR,f
  goto TEST_RUD										;counter still hasn't expired, move on
  movfw ESC_PIN_VAL
  xorwf PORTA,f
  andwf PORTA,w
  movlw cPFM_ON_TIME									;load PFM on time, Note that movlw doesn't affect Z
  btfsc STATUS,Z									;if Z is set, the pins are off, use the off value
  movfw ESC_PFM_VAL									;set cntr for desired time
  movwf ESC_PFM_CNTR

TEST_RUD
  btfss RUDTOGGLE
  return
  decfsz RUD_PFM_CNTR,f
  return											;counter still hasn't expired, move on
  movfw RUD_PIN_VAL
  xorwf PORTC,f
  andwf PORTC,w
  movlw cPFM_ON_TIME									;load PFM on time, Note that movlw doesn't affect Z
  btfsc STATUS,Z									;if Z is set, the pins are off, use the off value
  movfw RUD_PFM_VAL									;set cntr for desired time
  movwf RUD_PFM_CNTR

  return
;END DOPFM
;-------------------------------------

GETCHANHIGH
;lookup table for high byte of channels
  addwf PCL,f
  retlw HIGH(CHANNEL1)
  retlw HIGH(CHANNEL2)
  retlw HIGH(CHANNEL3)
  retlw HIGH(CHANNEL4)
  retlw HIGH(CHANNEL5)
  retlw HIGH(CHANNEL6)
  retlw HIGH(CHANNEL7)
  retlw HIGH(CHANNEL8)
;end GETCHANHIGH

GETCHANLOW
;lookup table for low byte of channels
  addwf PCL,f
  retlw LOW(CHANNEL1)
  retlw LOW(CHANNEL2)
  retlw LOW(CHANNEL3)
  retlw LOW(CHANNEL4)
  retlw LOW(CHANNEL5)
  retlw LOW(CHANNEL6)
  retlw LOW(CHANNEL7)
  retlw LOW(CHANNEL8)
;end GETCHANLOW


 END
 
