;DIY-UHF Digital signal Encoding TX
;David Lancaster
;All rights reserved

;History
;2.1 Moved to using 5Mhz clkout from IA4220 chip for more accurate bit timing
;2.2
;	13pf XTAL impedance to match CTX658
;	Cleanup sending code and HEADER/frame definition
;	Added ID1, ID2, FRAME0, removing header
;	Switched to only supporting 16F630 (more pins for ID code & binding)
;	Correct pinout in code
;	Added FHSS 
;	Restructured MAIN & SENDDATA
;	Replaced parity by checksum
;	Added 10pf XTAL option for CS5032
;	Switched to 10Mhz clock speed for greater accuracy
;	Rebranded as version 2 for simplicity
;3.1
;	Added dwell counter to reduce number of actual channel changes
;	Switched to 8-bit channel values
;	Removed FHSS code

;CONFIGURABLE OPTIONS
;*** YOU PROBABLY HAVE TO SET THESE CORRECTLY!!! ***

;ID Code
;Define a "unique" pair of ID codes for yourself, try to make it randomish
;e.g. don't use "123", "42", "69", etc...
;Valid values are 1-255 (tho I'd avoid extremes)
#DEFINE ID1	d'123'
#DEFINE	ID2	d'27'
  MESSG "Using defined IDs...hope you set them..."

;Transmitter type 
;uncomment the correct one, comment out the others
;#DEFINE JR
#DEFINE FUTABA
;#DEFINE SANWA

;Set Logic levels, comment for inverted logic (1 = 0V, 0 = 5+V)
;Futaba and Hitec are *not* logic high...
;#DEFINE LOGIC_HIGH

;END CONFIGURABLE OPTIONS





	
;some limits to the RC PPM signal, exceptions are treated as errors
;Note that the channel width does *not* include the pulse width
;these may need fine tuning for specific radios/brands, value in mS

#define		MAX_CHANNEL_WIDTH			d'1950'		;maximum width of a channel
#define		MIN_CHANNEL_WIDTH			d'1050'		;minimum width of a channel
#define		MIN_PULSE_WIDTH				d'300'		;minimum width of a pulse
#define		MAX_PULSE_WIDTH				d'550'		;maximum width of a pulse




#IFDEF __16F630
	MESSG "Assembling for 16F630"
	LIST	P=PIC16F630
	INCLUDE "P16F630.INC"
	__CONFIG _MCLRE_OFF & _CP_OFF & _WDT_OFF & _EC_OSC & _CPD_OFF & _PWRTE_ON & _BODEN
	;ERRORLEVEL -302					;turn off register not in bank0 warnings
#ENDIF


;DEFINE CONSTANTS
;----------------

#INCLUDE ../data_format.inc

#define		CLOCKSPEED				d'10'
#define		CLKSKEW					CLOCKSPEED/d'4'		;Fosc in Mhz/4 (1op per 4 osc)


#define		c3MS						d'3000'
#define		c17MS						d'17000'
#define		MAX_CHANNEL_WIDTH_TICKS		MAX_CHANNEL_WIDTH*CLKSKEW
#define		MIN_CHANNEL_WIDTH_TICKS		MIN_CHANNEL_WIDTH*CLKSKEW
#define		MIN_PULSE_WIDTH_TICKS		MIN_PULSE_WIDTH*CLKSKEW
#define		MAX_PULSE_WIDTH_TICKS		MAX_PULSE_WIDTH*CLKSKEW



#define		NUMERRORS				XMIT_BPS/d'140'	;about 1/2 second, give or take

;DIY-UHF TX 16F630 RX pinouts:
;
;		   	---------
;		Vcc	|1    14| Gnd
;CLK -->RA5	|2    13| RA0--> FSK
;nSEL<--RA4 |3    12| RA1--> SDI
;PPM -->RA3	|4    11| RA2--> SCK
;NC  <--RC5 |5    10| RC0<-- ALT_ID
;CB3 -->RC4 |6     9| RC1<-- BIND
;CB2 -->RC3 |7     8| RC2<-- CB1  (channel bit 1)
;		   	---------

;pin connections 
#define		IA_FSK		PORTA,0	
#define		IA_SDI		PORTA,1
#define		IA_SCK		PORTA,2
#define		INPUT_PIN	PORTA,3
#define		IA_nSEL		PORTA,4
#define		CLKIN		PORTA,5	;not an IO, just a placeholder

#define		USE_ALT_ID	PORTC,0	;use alternate id code
#define		ADV_BIND	PORTC,1	;advertise binding to RXes


;IA4220 settings
#define		IA_PWRMGT_OFF	b'1100000000000000'		;turn off units
#define		IA_PWRMGT_ON	b'1100000000111000'		;Manual control: Turn on crystal osc, synthesizer & power amp  & clock output, disable low-battery
;#define		IA_CONFIG		b'1001111110010010'		;915 Mhz band, 10Mhz clkout, 13pf XTAL impedance, Set Freq deviation to +/- 90 khz
#define		IA_CONFIG		b'1001111100110010'		;915 Mhz band, 10Mhz clkout, 10pf XTAL impedance, Set Freq deviation to +/- 90 khz
;#define		IA_FREQ			CHANNEL1				;Set to first channel (NOT USED ANYMORE)
#define		IA_PWR			b'10110000'				;FSK signalling, high power



;VARIABLES
;---------

#define CBVAL 0x020					;12F6XX/16F6XX regs start at 20h

 CBLOCK CBVAL						;Start a Variable block, starting at CBVAL set above
;Holds values received from TX
ESC_INPUT							;ESC buffer
AIL_INPUT							;AILERON buffer
ELE_INPUT							;Elevator buffer
RUD_INPUT							;Rudder buffer
CURRENTCHANNEL						;holds currentchannel-1 when reading input
CHANNELVALUE						;hold value of current channel when reading input

;registers to hold data to send
HEADER0								
HEADER1
HEADER2
HEADER3
FRAME0
FRAME1
FRAME2
FRAME3
FRAME4
FRAME5
FRAME6
TOTALBITCNTR						;number of bits to send

TEMP								;temp register

;used in SPI communication
IA_DATAL	
IA_DATAH	
SPI_BIT_COUNTER

;store ID we're actually using
ID_CODE
;PRNG for FHSS
LFSR

MISC								;misc flag holder
 ENDC

;usage of MISC reg
#define		BINDING					MISC, 0	;flag to indicate we're advertising binding to RX
#define		SEND_DATA_FLAG			MISC, 1	;Flag to indicate that we can send data


 ERRORLEVEL -207
;organize channels so that the values go where we expect them to
#IFDEF JR
;Config for JR, ESC, AIL, ELE, RUD
	CH1_INPUT	EQU ESC_INPUT
	CH2_INPUT	EQU AIL_INPUT
	CH3_INPUT	EQU ELE_INPUT
	CH4_INPUT	EQU RUD_INPUT
#ENDIF ;JR

#IFDEF FUTABA
;Config for Futaba/Hitec, AIL, ELE, ESC, RUD
	CH1_INPUT	EQU AIL_INPUT
	CH2_INPUT	EQU ELE_INPUT
	CH3_INPUT	EQU ESC_INPUT
	CH4_INPUT	EQU RUD_INPUT
#ENDIF ;FUTABA

#IFDEF SANWA
;Config for SANWA, AIL, ESC, ELE, RUD
	CH1_INPUT	EQU AIL_INPUT
	CH2_INPUT	EQU ESC_INPUT
	CH3_INPUT	EQU ELE_INPUT
	CH4_INPUT	EQU RUD_INPUT
#ENDIF ;SANWA


BIGWAIT MACRO
	;this is 65536 ticks (256*256)
  clrf TMR0									;TMR0 will overflow after 256 increments
  goBank1									;Set config options in Bank1
  bsf OPTION_REG,PS2						;Set TMR0 for 1:256 prescaler (111)
  bsf OPTION_REG,PS1
  bsf OPTION_REG,PS0
  goBank0
  btfss INTCON,T0IF	
  goto $-1
  bcf INTCON,T0IF
 ENDM

goBank1 MACRO
  bsf STATUS,RP0
  ERRORLEVEL -302	;disable bank warnings
 ENDM

goBank0 MACRO
  bcf STATUS,RP0
  ERRORLEVEL +302	;enable bank warnings
 ENDM


;Macros to simplify testing for input
#IFDEF LOGIC_HIGH
	;pulse defined as a digital 1
	btpulsesc macro
			btfsc INPUT_PIN	
	endm
	btpulsess macro
			btfss INPUT_PIN	
	endm					
#ELSE
	;pulse defined as a digital 0
	btpulsesc macro
			btfss INPUT_PIN	
	endm
	btpulsess macro
			btfsc INPUT_PIN	
	endm							
#ENDIF
 ERRORLEVEL +207





;PROGRAM CODE STARTS HERE
 ORG 0  
  Goto SETUP


;-------------------------------------
SENDDATA
;START SUBROUTINE SENDDATA
  movlw PREAMBLE		;load preamble into first two bytes
  movwf HEADER0
  movwf HEADER1
  movlw HIGH(SYNCHRON)	;load synchron into next two bytes
  movwf HEADER2
  movlw LOW(SYNCHRON)
  movwf HEADER3

  call WAITBIT_NOWAIT			;this call is to start TMR0 counting

SENDABIT 						;test data, and send appropriate bit
  btfsc HEADER0,7
  goto $+3
  bcf IA_FSK
  goto $+2
  bsf IA_FSK

	;shift entire packet one bit to the left
  bcf STATUS,C
  rlf FRAME6,f
  rlf FRAME5,f
  rlf FRAME4,f
  rlf FRAME3,f
  rlf FRAME2,f
  rlf FRAME1,f
  rlf FRAME0,f
  rlf HEADER3,f
  rlf HEADER2,f
  rlf HEADER1,f
  rlf HEADER0,f

  call WAITBIT					;wait a bit's duration

  decfsz TOTALBITCNTR,f
  goto SENDABIT

	;clear FSK pin
  bcf IA_FSK
  
  return

;END SENDDATA
;-------------------------------------



;-------------------------------------
WAITBIT
	;wait until TMR0 expires
  btfss INTCON,T0IF
  goto $-1
WAITBIT_NOWAIT	;entry point without corresponding wait, for setup only
	;reset TMR0 & clear interrupt flag (we're not using interrupts, but the flag still works
  bcf INTCON,T0IF
  movlw d'256'-(XMIT_uSPB*CLKSKEW)+7		;***requries hand-tuning***
  movwf TMR0
  goBank1									;Set config options in Bank1
  movlw b'00001000'							;TMR0 config: prescaler assigned to WDT, internal clocking 
  movwf OPTION_REG
  goBank0

  return
;END WAIT
;-------------------------------------



;-------------------------------------
READ_INPUT
;START SUBROUTINE READ_INPUT
	;*****************************************
		;read input from RC TX, parsing out the values for the first four channels
		;uses some hard coded max/mins & some defines
	;*****************************************
	;Loop until we have no pulses for at least 3ms, using TMR0
  bcf INTCON,T0IF
  movlw d'256'-((c3MS*CLKSKEW)/d'32')		;calculate 3ms in TMR0 ticks, prescaled by 1:32
  movwf TMR0
  movlw b'00000100'							;prescaled by 1:32
  goBank1									;Set config options in Bank1
  movwf OPTION_REG							;prescale tmr0
  goBank0
WAIT3ms
  btfsc INTCON,T0IF
  goto STARTREADING
  btpulsess
  goto WAIT3ms
  goto READ_INPUT		;hrm, technically, we can get stuck in here with no signal, but who cares

STARTREADING
	;Wait for a pulse in the next 17ms, using TMR0.  Error out after 17ms
  bcf INTCON,T0IF
  movlw d'256'-((c17MS*CLKSKEW)/d'256')		;calculate 17ms in TMR0 ticks, prescaled by 1:256
  movwf TMR0
  movlw b'00000111'							;prescaled by 1:256
  goBank1									;Set config options in Bank1
  movwf OPTION_REG							;prescale tmr0
  goBank0
WAIT17ms
  btfsc INTCON,T0IF
  goto INPUT_ERROR
  btpulsess
  goto WAIT17ms


;Start reading train of channels
	;set channel counter to first channel (n-1)
  movlw d'0'
  movwf CURRENTCHANNEL

READPULSE
	;We count the channel on TMR0, treating > MAX_CHANNEL_WIDTH as an error
	;start our channel timer here
  bcf INTCON,T0IF
  movlw d'256'-((MIN_CHANNEL_WIDTH_TICKS)/d'16' + d'5')	;load the min channel width
  movwf TMR0
  movlw b'00000011'							;prescale by 1:16
  goBank1									;Set config options in Bank1
  movwf OPTION_REG							;prescale tmr0
  goBank0
WAITONPULSE
  btpulsesc									;wait for the pulse to expire
  goto $-1
WAITMINWIDTH
  btpulsesc				;spin until we exhaust the min channel width or we see a pulse
  goto MINWIDTH
  btfss INTCON,T0IF							
  goto WAITMINWIDTH
MINWIDTH
  clrf TMR0
  bcf INTCON,T0IF
  movlw b'00000010'							;prescale by 1:8
  goBank1									;Set config options in Bank1
  movwf OPTION_REG							;prescale tmr0
  goBank0
WAITONCHAN
  btpulsess
  goto WAITONCHAN
	;now we have a value for the channel, quickly save it and start the timer for the next channel
  movfw TMR0
  btfsc INTCON,T0IF							;check if we overflowed
  movlw d'255'								;if so, set it to maximum value
  movwf CHANNELVALUE
	;set TMR0 so it starts counting from *NOW* (which is actually a few us late)
  bcf INTCON,T0IF
  movlw d'256'-((MIN_CHANNEL_WIDTH_TICKS)/d'16' + d'5')	;load the min channel width
  movwf TMR0
  movlw b'00000011'
  goBank1									;Set config options in Bank1
  movwf OPTION_REG							;prescale tmr0 by 1:16
  goBank0

	;Save the value
  movfw CURRENTCHANNEL
  incf CURRENTCHANNEL,f
  addwf PCL,f
  goto CH1
  goto CH2
  goto CH3
  goto CH4

CH1
  movfw CHANNELVALUE
  movwf CH1_INPUT
  goto WAITONPULSE
CH2
  movfw CHANNELVALUE
  movwf CH2_INPUT
  goto WAITONPULSE
CH3
  movfw CHANNELVALUE
  movwf CH3_INPUT
  goto WAITONPULSE
CH4
  movfw CHANNELVALUE
  movwf CH4_INPUT

	;done reading channels, set some flags
  bsf SEND_DATA_FLAG			;assert that we want to be sending data these days
  return

INPUT_ERROR
	;Error reading input
	;Pin exceeded timeout or pulse width, etc
  return

;-------------------------------------
;END READ_INPUT


;-------------------------------------
;Sub send_SPI
send_SPI
	;Sub to send specified # of bits (in W) via SPI (up to 16) from registers IA_DATAL & IA_DATAH
	;Starts with MSB of IA_DATAH, so bits should be left justified to IA_DATAH
  movwf SPI_BIT_COUNTER
  bcf IA_SCK
  bcf IA_SDI
  bcf IA_nSEL	;nSEL set low to open SPI conversation
  goto $+1
send_SPI_BIT
  bcf IA_SDI	;set value on SD output (SDI on IA chip)
  btfsc IA_DATAH,7
  bsf IA_SDI
  goto $+1
  bsf IA_SCK	;SCK rising edge
  goto $+1
  rlf IA_DATAL,f
  rlf IA_DATAH,f
  bcf IA_SCK	;SCK falling edge
  decfsz SPI_BIT_COUNTER,f
  goto send_SPI_BIT
  goto $+1
  bsf IA_nSEL	;finish SPI conversation
  goto $+1
  bcf IA_SDI
  return
;END send_SPI
;-------------------------------------




;-------------------------------------

SETUP
	;NOTE: Since we're using the IA4220 for clock, note that it starts at 1Mhz, which we later change
;PIC SETUP
  clrf PORTA 								;clear outputs
  clrf PORTC								;clear outputs

  goBank1									;Set config options in Bank1
  call 3FFh									
  movwf OSCCAL 								;config OSCCAL
  movlw b'00000000'							;TMR0 config at default
  movwf OPTION_REG
  movlw b'00001000'							;set GP3 to input, GP0-2,4,5 as outputs
  movwf TRISA
  movlw b'00011111'							;RC0-RC4 as inputs	
  movwf TRISC
  goBank0									;done with Bank1 config options
  movlw 07h									;Turn off Comparator unit (111)
  movwf CMCON

;done PIC setup


;init protocol

  bcf SEND_DATA_FLAG						;make sure we don't suffer premature sendulation

;choose and load ID_CODE with either ID1 or ID2, depending on input pin
  movlw ID1
  btfsc USE_ALT_ID
  movlw ID2
  movwf ID_CODE

;prime LFSR using ID_CODE
  movfw ID_CODE
  movwf LFSR
  

;done protocol setup


;do IA422X setup
  bsf IA_nSEL		;it's logic low, so set it high

;sleep for a bit so the TX chip can stabilize after it's POR
  movlw d'250000'/d'65536'					;one second, rounded down
  movwf TEMP
OUTER_LOOP
  BIGWAIT	;Note that at the startup 1Mhz clock rate, it's actually five times as long as later
  decfsz TEMP,f
  goto OUTER_LOOP

;these require checking the IA422X datasheet, have fun
	;Power Management Setting *shut down tx first*
  movlw HIGH(IA_PWRMGT_OFF)
  movwf IA_DATAH
  movlw LOW(IA_PWRMGT_OFF)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  BIGWAIT
	;Configuration Setting
  movlw HIGH(IA_CONFIG)
  movwf IA_DATAH
  movlw LOW(IA_CONFIG)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI 
	;Frequency Setting
  movfw PORTC			;pull in setting from PORTC switches
  andlw b'00011100'		;mask out the other bits
  movwf TEMP
  bcf STATUS,C			;and shift two bits to the right
  rrf TEMP,f
  rrf TEMP,f
  pageselw GETCHANHIGH
  movfw TEMP
  call GETCHANHIGH		;lookup high byte of channel
  movwf IA_DATAH	
  pageselw GETCHANLOW
  movfw TEMP
  call GETCHANLOW		;lookup low byte of channel
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
  pageselw SETUP		;reset the PCLATH bits
	;Power Setting
  movlw IA_PWR
  movwf IA_DATAH
  movlw d'8'
  call send_SPI
	;Power Management Setting
  movlw HIGH(IA_PWRMGT_ON)
  movwf IA_DATAH
  movlw LOW(IA_PWRMGT_ON)
  movwf IA_DATAL
  movlw d'16'
  call send_SPI
	;No changes to 	Bit Rate, Low Battery, Sleep, push-button, or wake-up timer config
	;Wait for Oscillator & PLL to stabilize
  nop
  BIGWAIT

MAIN
	;main body of the code
	;we loop in here forever
  call READ_INPUT

  btfss SEND_DATA_FLAG		;only send data if the flag is set
  goto MAIN	

  ;move latest values to output buffers, checksum and send
  clrf CHECKSUM_VAL
  movlw BITSPERPACKET
  movwf TOTALBITCNTR

  movfw ID_CODE
  movwf ID_CODE_VAL
  addwf CHECKSUM_VAL,f

  movlw REG_TX				;flag as normal transmission
  btfsc ADV_BIND			;unless the binding pin is set
  movlw ADV_BINDING			; in which case it's a binding transmission
  movwf SIGOPTS_VAL
  addwf CHECKSUM_VAL,f

  movfw AIL_INPUT
  movwf AIL_VAL
  addwf CHECKSUM_VAL,f

  movfw ELE_INPUT
  movwf ELE_VAL
  addwf CHECKSUM_VAL,f

  movfw ESC_INPUT
  movwf ESC_VAL
  addwf CHECKSUM_VAL,f

  movfw RUD_INPUT
  movwf RUD_VAL
  addwf CHECKSUM_VAL,f

  call SENDDATA				
  goto MAIN
;END of main program, loop back
;-------------------------------------


GETCHANHIGH
;lookup table for high byte of channels
  addwf PCL,f
  retlw HIGH(CHANNEL1)
  retlw HIGH(CHANNEL2)
  retlw HIGH(CHANNEL3)
  retlw HIGH(CHANNEL4)
  retlw HIGH(CHANNEL5)
  retlw HIGH(CHANNEL6)
  retlw HIGH(CHANNEL7)
  retlw HIGH(CHANNEL8)
;end GETCHANHIGH

GETCHANLOW
;lookup table for low byte of channels
  addwf PCL,f
  retlw LOW(CHANNEL1)
  retlw LOW(CHANNEL2)
  retlw LOW(CHANNEL3)
  retlw LOW(CHANNEL4)
  retlw LOW(CHANNEL5)
  retlw LOW(CHANNEL6)
  retlw LOW(CHANNEL7)
  retlw LOW(CHANNEL8)
;end GETCHANLOW

  END


